
/*
	roi.c:	Functions revolving around the region of interest 
*/


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "roi.h"
#include "interface.h"
#include "application.h"
#include "functions.h"


static char rectangle_ini = 0;
static int lastx1, lasty1, lastx2, lasty2, lastox, lastoy;
static int marchingant_offset = 0;
roi_type dragroi;

/****************************************************************/

/* imgxy returns coordinates in the actual image from 
   display window coordinates */

void roi_imgxy (int* x, int* y)
{
int ix,iy;

	ix = mainimg.drawable->allocation.width;	/* size of the window */
	iy = mainimg.drawable->allocation.height;

	/* Convert mouse coords relative to window into image coords */
	/* we HAVE to use (w,h) instead of (im->width,im_height), because
	   apparently im->... is adjusted with the display size */


	ix = (int)((double)(*x) * (double)(mainimg.width)/(double)ix) * pow(2.0, -zoomlevel_x);
	iy = (int)((double)(*y) * (double)(mainimg.height)/(double)iy) * pow(2.0, -zoomlevel_y);

	*x=ix; *y=iy;
}


/* This one does just the inverse - converts image coords into dspy coords */

void roi_inv_imgxy (int* x, int* y)
{
int ix,iy;

	ix = mainimg.drawable->allocation.width;
	iy = mainimg.drawable->allocation.height;

	if (ix>mainimg.width)		
		ix = (int)(0.5+(double)(*x) * pow(2.0,zoomlevel_x) * (double)(ix)/(double)mainimg.width);
	else
		ix = (int)((double)(*x) * pow(2.0,zoomlevel_y) * (double)(ix)/(double)mainimg.width);

	if (iy>mainimg.height)
		iy = (int)(0.5+(double)(*y) * pow(2.0,zoomlevel_x) * (double)(iy)/(double)mainimg.height);
	else
		iy = (int)((double)(*y) * pow(2.0,zoomlevel_y) * (double)(iy)/(double)mainimg.height);
		
	*x=ix; *y=iy;
}



/* This function XORs a rectangle into the image. It also erases the
   previously drawn rectangle, thus allows to drag-draw it */

void roi_drag_rectangle  (int x1, int y1, int x2, int y2)
{

GdkGC *gc;
int i;

	gc = gdk_gc_new (mainimg.drawable->window);	/* allocate mem & assign null values */
	gdk_gc_copy (gc, mainimg.drawable->style->white_gc); /* Copy some values */
	gdk_gc_set_function (gc, GDK_XOR);

	if (rectangle_ini)	/* we'll have to erase the old rectangle */
	{
		gdk_draw_rectangle (mainimg.drawable->window, gc, FALSE,
							lastx1,lasty1,lastx2-lastx1,lasty2-lasty1);
	}
	
	/* Make sure that x2-x1 and y2-y1 is always positive */
	if (x2<x1) { i=x1; x1=x2; x2=i; }
	if (y2<y1) { i=y1; y1=y2; y2=i; }
	
	gdk_draw_rectangle (mainimg.drawable->window, gc, FALSE, x1,y1,x2-x1,y2-y1);
	rectangle_ini = 1;
	lastx1=x1;
	lasty1=y1;
	lastx2=x2;
	lasty2=y2;
	gdk_gc_destroy (gc);	/* free the memory associated with the gc */
}


/* End dragging - erase the tentative rectangle */

void roi_erase_rectangle ()
{

GdkGC *gc;


	if (rectangle_ini)	/* we'll have to erase the old rectangle */
	{
		gc = gdk_gc_new (mainimg.drawable->window);	/* allocate mem & assign null values */
		gdk_gc_copy (gc, mainimg.drawable->style->white_gc); /* Copy some values */
		gdk_gc_set_function (gc, GDK_XOR);
		gdk_draw_rectangle (mainimg.drawable->window, gc, FALSE,
							lastx1,lasty1,lastx2-lastx1,lasty2-lasty1);
		gdk_gc_destroy (gc);	/* free the memory associated with the gc */
	}
	
	rectangle_ini = 0;
}


/* same as above, but drags a general polygonal ROI */


void roi_drag_roi  (int ofs_x1, int ofs_y1)
{

GdkGC *gc;
int i,j,x,y,x1,y1,x2,y2;

	if (current_roi.roi_mode==ROI_NONE) return;	/* Only when a ROI has been defined */
	
	gc = gdk_gc_new (mainimg.drawable->window);	/* allocate mem & assign null values */
	gdk_gc_copy (gc, mainimg.drawable->style->white_gc); /* Copy some values */
	gdk_gc_set_function (gc, GDK_XOR);

	for (j=!dragroi_ini; j<=1; j++)
	{
		if (j==0)					/* Erase old ROI */
		{
			x=lastox; y=lastoy;
		}
		else
		{
			x=ofs_x1; y=ofs_y1;
		}
			

		if (current_roi.roi_mode==ROI_RECT)
		{
			x1=x+current_roi.xi[0]; x2=x+current_roi.xi[1];
			y1=y+current_roi.yi[0]; y2=y+current_roi.yi[1];
			roi_inv_imgxy (&x1,&y1); roi_inv_imgxy (&x2,&y2);
			gdk_draw_rectangle (mainimg.drawable->window, gc, FALSE, x1,y1,x2-x1,y2-y1);
		}
		
		if ((current_roi.roi_mode==ROI_POLY) || (current_roi.roi_mode==ROI_CIRCLE))
		{
			x1=x+current_roi.xi[0]; y1=y+current_roi.yi[0];
			roi_inv_imgxy (&x1,&y1);

			for (i=1; i<current_roi.numvertices; i++)
			{
				x2=x+current_roi.xi[i]; y2=y+current_roi.yi[i];
				roi_inv_imgxy (&x2,&y2);
				gdk_draw_line (mainimg.drawable->window, gc, x1,y1,x2,y2);
				x1=x2; y1=y2;
			}
						
		}
	}

	dragroi_ini = 1;
	lastox=ofs_x1;
	lastoy=ofs_y1;
	gdk_gc_destroy (gc);	/* free the memory associated with the gc */
}



void roi_drag_vertex  (int vertexno, int ofs_x1, int ofs_y1)
{

	GdkGC *gc;
	int i,j,x,y,x1,y1,x2,y2;

	if (current_roi.roi_mode!=ROI_POLY) return;		/* Only polygonal ROI */
	
	gc = gdk_gc_new (mainimg.drawable->window);	/* allocate mem & assign null values */
	gdk_gc_copy (gc, mainimg.drawable->style->white_gc); /* Copy some values */
	gdk_gc_set_function (gc, GDK_XOR);

	for (j=!dragroi_ini; j<=1; j++)	/* Loop: First erase old ROI, then draw new one */
	{
		if (j==0)					/* Erase old ROI */
		{
			x=lastox; y=lastoy;
		}
		else
		{
			x=ofs_x1; y=ofs_y1;
		}
			

		for (i=0; i<current_roi.numvertices; i++)
		{
			if ( (i==vertexno-1) || ((vertexno==1) && (i==current_roi.numvertices-1)) )
			{
				x2=x+current_roi.xi[i]; y2=y+current_roi.yi[i];
			}
			else
			{
				x2=current_roi.xi[i]; y2=current_roi.yi[i];	/* Unchanged vertices */
			}
			roi_inv_imgxy (&x2,&y2);
			if (i>0) gdk_draw_line (mainimg.drawable->window, gc, x1,y1,x2,y2);
			x1=x2; y1=y2;
		}

	}

	dragroi_ini = 1;
	lastox=ofs_x1;
	lastoy=ofs_y1;
	gdk_gc_destroy (gc);	/* free the memory associated with the gc */
}


void roi_erase_roi  ()
{

GdkGC *gc;
int i,x,y,x1,y1,x2,y2;

	if (dragroi_ini==0) return;	/* Only when a ROI has been dragged */
	
	gc = gdk_gc_new (mainimg.drawable->window);	/* allocate mem & assign null values */
	gdk_gc_copy (gc, mainimg.drawable->style->white_gc); /* Copy some values */
	gdk_gc_set_function (gc, GDK_XOR);

	x=lastox; y=lastoy;

	if (current_roi.roi_mode==ROI_RECT)
	{
		x1=x+current_roi.xi[0]; x2=x+current_roi.xi[1];
		y1=y+current_roi.yi[0]; y2=y+current_roi.yi[1];
		roi_inv_imgxy (&x1,&y1); roi_inv_imgxy (&x2,&y2);
		gdk_draw_rectangle (mainimg.drawable->window, gc, FALSE, x1,y1,x2-x1,y2-y1);
	}

	if ((current_roi.roi_mode==ROI_POLY) || (current_roi.roi_mode==ROI_CIRCLE))
	{
		x1=x+current_roi.xi[0]; y1=y+current_roi.yi[0];
		roi_inv_imgxy (&x1,&y1);

		for (i=1; i<current_roi.numvertices; i++)
		{
			x2=x+current_roi.xi[i]; y2=y+current_roi.yi[i];
			roi_inv_imgxy (&x2,&y2);
			gdk_draw_line (mainimg.drawable->window, gc, x1,y1,x2,y2);
			x1=x2; y1=y2;
		}

	}

	dragroi_ini = 0;
	gdk_gc_destroy (gc);	/* free the memory associated with the gc */
}


/* Add the offset to the current ROI */

void roi_offset_roi  (int x, int y)
{
int i,x1,y1,x2,y2;
int boundflg;			/* If we run into an image bound */

	boundflg=0;
	if (current_roi.roi_mode==ROI_RECT)
	{
		x1=x+current_roi.xi[0]; x2=x+current_roi.xi[1];
		y1=y+current_roi.yi[0]; y2=y+current_roi.yi[1];
		if (x1<0) x1=0; else if (x1>=mainimg.xmax) x1=mainimg.xmax;
		if (y1<0) y1=0; else if (y1>=mainimg.ymax) y1=mainimg.ymax;
		if (x2<0) x2=0; else if (x2>=mainimg.xmax) x2=mainimg.xmax;
		if (y2<0) y2=0; else if (y2>=mainimg.ymax) y2=mainimg.ymax;
		current_roi.xi[0]=x1; 		current_roi.yi[0]=y1;
		current_roi.xi[1]=x2; 		current_roi.yi[1]=y2;
		current_roi.bbox_x1 =x1;	current_roi.bbox_y1 =y1;
		current_roi.bbox_x2 =x2;	current_roi.bbox_y2 =y2;
	}
	if ((current_roi.roi_mode==ROI_POLY) || (current_roi.roi_mode==ROI_CIRCLE))
	{
		for (i=0; i<current_roi.numvertices; i++)
		{
			x1=x+current_roi.xi[i]; y1=y+current_roi.yi[i];
			if (x1<0) x1=0; else if (x1>=mainimg.xmax) { x1=mainimg.xmax; boundflg=1; }
			if (y1<0) y1=0; else if (y1>=mainimg.ymax) { y1=mainimg.ymax; boundflg=1; }
			current_roi.xi[i]=x1; current_roi.yi[i]=y1;
			
			/* If it is circular and we need to adjust out-of-bound vertices, convert to polygonal */
			
			if ((current_roi.roi_mode==ROI_CIRCLE) && (boundflg)) current_roi.roi_mode=ROI_POLY;
			
		}
		if (current_roi.roi_mode==ROI_CIRCLE)
		{
			current_roi.xc += x;
			current_roi.yc += y;
		}

		/* Re-create the bounding box by scanning over the vertices */
		
		current_roi.bbox_x1 = current_roi.bbox_x2 = current_roi.xi[0];
		current_roi.bbox_y1 = current_roi.bbox_y2 = current_roi.yi[0];
		for (i=1; i<current_roi.numvertices; i++)
		{
			if (current_roi.xi[i] < current_roi.bbox_x1) current_roi.bbox_x1=current_roi.xi[i];
			if (current_roi.xi[i] > current_roi.bbox_x2) current_roi.bbox_x2=current_roi.xi[i];
			if (current_roi.yi[i] < current_roi.bbox_y1) current_roi.bbox_y1=current_roi.yi[i];
			if (current_roi.yi[i] > current_roi.bbox_y2) current_roi.bbox_y2=current_roi.yi[i];
		}		
	}

}


void roi_offset_vertex (int vertexno, int x, int y)
{
int i;

	if (current_roi.roi_mode!=ROI_POLY) return;		/* Only allowed in Poly mode */
	roi_imgxy (&x,&y);		/* Mouse to image coordinates */
	
	current_roi.xi[vertexno-1]+=x;
	current_roi.yi[vertexno-1]+=y;
	if (vertexno=1)
	{
		current_roi.xi[current_roi.numvertices]=current_roi.xi[0];
		current_roi.yi[current_roi.numvertices]=current_roi.yi[0];
		current_roi.xi[current_roi.numvertices-1]=current_roi.xi[0];
		current_roi.yi[current_roi.numvertices-1]=current_roi.yi[0];
	}
	
	/* Re-create the bounding box by scanning over the vertices */
	
	current_roi.bbox_x1 = current_roi.bbox_x2 = current_roi.xi[0];
	current_roi.bbox_y1 = current_roi.bbox_y2 = current_roi.yi[0];
	for (i=1; i<current_roi.numvertices; i++)
	{
		if (current_roi.xi[i] < current_roi.bbox_x1) current_roi.bbox_x1=current_roi.xi[i];
		if (current_roi.xi[i] > current_roi.bbox_x2) current_roi.bbox_x2=current_roi.xi[i];
		if (current_roi.yi[i] < current_roi.bbox_y1) current_roi.bbox_y1=current_roi.yi[i];
		if (current_roi.yi[i] > current_roi.bbox_y2) current_roi.bbox_y2=current_roi.yi[i];
	}		
}




/*===========================================================================*/



void kill_roi ()
{
char save_roi;

	save_roi = current_roi.roi_mode;
	current_roi.roi_mode=0;
	current_roi.numvertices=0;
	display_clickstack=0;

	if (save_roi)	/* Old ROI specified, so redraw just in case */
	{
		redraw (0);
	}
	
}


/* Specify a rectangular ROI */

void roi_set_rectangular (int x1, int y1, int x2, int y2)
{
int i;

	if (current_roi.roi_mode)	/* Old ROI specified, so erase */
	{
		redraw (0);
	}
	
	/* Make sure that x2-x1 and y2-y1 is always positive */
	if (x2<x1) { i=x1; x1=x2; x2=i; }
	if (y2<y1) { i=y1; y1=y2; y2=i; }

	/* Convert click coordinates into image coords */

	roi_imgxy (&x1,&y1);
	roi_imgxy (&x2,&y2);

	current_roi.xi[0]=x1;
	current_roi.xi[1]=x2;
	current_roi.xi[2]=-1;
	current_roi.yi[0]=y1;
	current_roi.yi[1]=y2;
	current_roi.yi[2]=-1;

	current_roi.bbox_x1 = x1;	/* Bounding box is the same */
	current_roi.bbox_x2 = x2;
	current_roi.bbox_y1 = y1;
	current_roi.bbox_y2 = y2;

	current_roi.roi_mode = ROI_RECT;	/* rectangular ROI - enables marching ants */
	marchingant_offset=0;
	dp (1, "ROI: Set new rectangular roi\n");


}


/* Polygonal ROIs: add a new vertex (used by the
   shift-click handler for the display window)
   The end of a polygon is indicated by a coord pair 
   of (-1,-1) as well as by numvertices */

void roi_add_vertex (int x1, int y1)
{
GdkGC *gc;

	if (current_roi.roi_mode==ROI_RECT) kill_roi();
	if ((current_roi.roi_mode==ROI_POLY) && roi_is_closed_polygon()) kill_roi();
	if (current_roi.roi_mode==ROI_CIRCLE) kill_roi();
	
	if (current_roi.numvertices<ROI_MAXVERTICES-3)
	{
		gc = gdk_gc_new (mainimg.drawable->window);	/* allocate mem & assign null values */
		gdk_gc_copy (gc, mainimg.drawable->style->white_gc); /* Copy some values */
		gdk_gc_set_function (gc, GDK_XOR);

		current_roi.xi[current_roi.numvertices]=x1;
		current_roi.yi[current_roi.numvertices]=y1;
		current_roi.numvertices++;
		current_roi.xi[current_roi.numvertices]=-1;
		current_roi.yi[current_roi.numvertices]=-1;

		if (current_roi.numvertices>1)
		{
			gdk_draw_line (mainimg.drawable->window, gc, 
							current_roi.xi[current_roi.numvertices-2],
							current_roi.yi[current_roi.numvertices-2],
							x1,y1);
		}

		gdk_gc_destroy (gc);	/* free the memory associated with the gc */
	}	

}

/* The last step in creating a polygonal ROI is closing it. 
   This means that the last vertex is identical to the first.
   Important: Due to the above restriction, ROI_MAXVERTICES-3,
   we should always have one entry left at this point! */

void roi_close_polygon (int convert)
{
int i;

	if (current_roi.numvertices<3) return;
	current_roi.xi[current_roi.numvertices]=current_roi.xi[0];
	current_roi.yi[current_roi.numvertices]=current_roi.yi[0];
	current_roi.numvertices++;
	current_roi.xi[current_roi.numvertices]=-1;
	current_roi.yi[current_roi.numvertices]=-1;
	
	/* Convert all vertices into image coordinates */
	
	if (convert)
		for (i=0; i<current_roi.numvertices; i++)
			roi_imgxy (&current_roi.xi[i], &current_roi.yi[i]);

	/* Determine the bounding box */
	
	current_roi.bbox_x1 = current_roi.bbox_x2 = current_roi.xi[0];
	current_roi.bbox_y1 = current_roi.bbox_y2 = current_roi.yi[0];
	for (i=1; i<current_roi.numvertices; i++)
	{
		if (current_roi.bbox_x1 > current_roi.xi[i]) current_roi.bbox_x1 = current_roi.xi[i];
		if (current_roi.bbox_x2 < current_roi.xi[i]) current_roi.bbox_x2 = current_roi.xi[i];
		if (current_roi.bbox_y1 > current_roi.yi[i]) current_roi.bbox_y1 = current_roi.yi[i];
		if (current_roi.bbox_y2 < current_roi.yi[i]) current_roi.bbox_y2 = current_roi.yi[i];
	}
	
	/* Set off the marching ants */

	marchingant_offset=0;
	current_roi.roi_mode=ROI_POLY;
	dp (1, "ROI: Set new polygonal roi\n");
	

}


/* Checks if polygon drawing is finished (i.e. we have a closed poly) 
   by comparing the first and last vertex */

char roi_is_closed_polygon()
{

	if (current_roi.roi_mode!=ROI_POLY) return 0;	/* only in polygon mode */
	if ((current_roi.xi[0]==current_roi.xi[current_roi.numvertices-1])
		&& (current_roi.yi[0]==current_roi.yi[current_roi.numvertices-1]))
	{
		return 1;
	}
	return 0;
}


/* Define a polygonal ROI that approximates a circle
   NOTE! This will always use image coordinates and not mouse coords */

void roi_set_circular_roi (int xc, int yc, int radius)
{
int i,n;

	if (current_roi.roi_mode==ROI_RECT) kill_roi();
	if ((current_roi.roi_mode==ROI_POLY) && roi_is_closed_polygon()) kill_roi();
	if (current_roi.roi_mode==ROI_CIRCLE) kill_roi();

	n=16;	/* 16 vertices */
	
	for (i=0; i<=n; i++)
	{
		current_roi.xi[i] = (int)(0.5 + xc + radius*cos(2*PI*i/n));
		current_roi.yi[i] = (int)(0.5 + yc + radius*sin(2*PI*i/n));
	}
	current_roi.numvertices=n+1;
	current_roi.xi[current_roi.numvertices]=-1;
	current_roi.yi[current_roi.numvertices]=-1;
	
	/* Set the bounding box */
	
	current_roi.bbox_x1 = xc-radius;
	current_roi.bbox_x2 = xc+radius;
	current_roi.bbox_y1 = yc-radius;
	current_roi.bbox_y2 = yc+radius;

	/* Set off the marching ants */

	marchingant_offset=0;
	current_roi.roi_mode=ROI_CIRCLE; current_roi.xc = xc; current_roi.yc = yc;
	current_roi.radius=radius;
	dp (1, "ROI: Set new circular roi with center %d,%d and radius %d\n",xc,yc,radius);

}


int roi_is_near_vertex (int xmouse, int ymouse)
{
int x,y,i;

	if (current_roi.roi_mode!=ROI_POLY) return 0;		/* Only valid in polygonal mode */

	x=xmouse; y=ymouse; roi_imgxy (&x, &y);	/* Mouse to image coordinates */


	for (i=0; i<current_roi.numvertices; i++)
	{
		if (( SQR(x-current_roi.xi[i])+ SQR(y-current_roi.yi[i]) ) <= 16)
			return i+1;		/* Return nonzero, exactly: index plus 1*/

	}
	return 0;		/* No vertex in proximity */
}




/* Timer handler that allows the marching ants to move */

gboolean roi_marchingants (gpointer data)
{
GdkGC *gc;
int x1,y1,x2,y2,i;

	if (mainimg.drawable==NULL) return TRUE;
	if (mainimg.drawable->window==NULL) return TRUE;	/* no window open */
	
	roicnt++; if (roicnt>100) roicnt=0;
	marchingant_offset +=3; if (marchingant_offset>9) marchingant_offset=0;

	if (current_roi.roi_mode)
	{
		gc = gdk_gc_new (mainimg.drawable->window);	/* allocate mem & assign null values */
		gdk_gc_copy (gc, mainimg.drawable->style->white_gc); /* Copy some values */
		gdk_gc_set_function (gc, GDK_XOR);
		gdk_gc_set_dashes (gc, marchingant_offset, "\3\3", 2);
		
		gdk_gc_set_line_attributes (gc,1,GDK_LINE_ON_OFF_DASH,GDK_CAP_NOT_LAST,GDK_JOIN_MITER);
		if (current_roi.roi_mode==ROI_RECT)
		{
			x1=current_roi.xi[0]; x2=current_roi.xi[1];
			y1=current_roi.yi[0]; y2=current_roi.yi[1];
			roi_inv_imgxy (&x1,&y1); roi_inv_imgxy (&x2,&y2);
			gdk_draw_rectangle (mainimg.drawable->window, gc, FALSE, x1,y1,x2-x1,y2-y1);
		}
		
		if ((current_roi.roi_mode==ROI_POLY) || (current_roi.roi_mode==ROI_CIRCLE))
		{
			x1=current_roi.xi[0]; y1=current_roi.yi[0];
			roi_inv_imgxy (&x1,&y1);

			for (i=1; i<current_roi.numvertices; i++)
			{
				x2=current_roi.xi[i]; y2=current_roi.yi[i];
				roi_inv_imgxy (&x2,&y2);
				gdk_draw_line (mainimg.drawable->window, gc, x1,y1,x2,y2);
				x1=x2; y1=y2;
			}
						
		}
		gdk_gc_destroy (gc);					
	}
	
	if ((display_clickstack) && (clickstack_last>=display_clickstack))
	{
		gc = gdk_gc_new (mainimg.drawable->window);	/* allocate mem & assign null values */
		gdk_gc_copy (gc, mainimg.drawable->style->white_gc); /* Copy some values */
		gdk_gc_set_function (gc, GDK_XOR);
		gdk_gc_set_dashes (gc, marchingant_offset, "\1\3", 2);
		gdk_gc_set_line_attributes (gc,1,GDK_LINE_ON_OFF_DASH,GDK_CAP_NOT_LAST,GDK_JOIN_MITER);
	
		for (i=0; i<display_clickstack-1; i++)
		{
			x1=clickstack[clickstack_last-i-1].x; x2=clickstack[clickstack_last-i-2].x;
			y1=clickstack[clickstack_last-i-1].y; y2=clickstack[clickstack_last-i-2].y;
			roi_inv_imgxy (&x1,&y1); roi_inv_imgxy (&x2,&y2);
			gdk_draw_line (mainimg.drawable->window, gc, x1,y1,x2,y2);
		}
	}
	
	return TRUE;	/* to keep it going */
	
}


/*==============================================================

	Check whether a point (x,y) is within the ROI
	Return 1 if yes, 0 otherwise 
	If no ROI, always return 1

================================================================*/


/* 	The winding-number algorithm was taken from 
	http://geometryalgorithms.com/Archive/algorithm_0103/algorithm_0103.htm

	roi_is_left tests if point (x2,y2) is Left|On|Right 
	of the infinite line running through v0:(x0,y0) to v1:(x1,y1)
	returns: >0 for left, 0 for on, and <0 for 
	right of the line.
	
	The crossing number algorithm and the winding number algorithm
	(both inplemented here) are slightly different when it comes to
	self-overlapping polygons. Their efficiency is about the same,
	but the winding-number algorithm is slightly more intuitive.
	
	crossing_number returns the number of times a ray starting from a
	point x1,y1 crosses a polygon boundary edge separating
	it's inside and outside.� If this number is even, then the point is
	outside; otherwise, when the crossing number is odd, 
	the point is inside.

	winding_number returns the number of times the polygon
	winds around (x1,y1). If the result is ==0, the point 
	lies outside of the polygon
*/

int roi_is_left (int v0, int v1, int x2,int y2)
{
int x0,y0,x1,y1;

	x0 = current_roi.xi[v0];
	y0 = current_roi.yi[v0];
	x1 = current_roi.xi[v1];
	y1 = current_roi.yi[v1];

	return ( (x1 - x0) * (y2 - y0) - (x2 - x0) * (y1 - y0) );

}


int winding_number (int x1, int y1)
{
int n; 		/* the crossing / winding number counter */
int i;



#ifdef crossing_number
	
	/* just lazy typing */
	#define C current_roi

	/* The crossing number algorithm */

	n=0;
	for (i=0; i<current_roi.numvertices-1; i++)
	{
		if (((C.yi[i] <= y1) && (C.yi[i+1] > y1))		/* an upward crossing */
			|| ((C.yi[i] > y1) && (C.yi[i+1] <= y1)))	/* a downward crossing */
		{
			float vt = (float)(y1 - C.yi[i]) / (C.yi[i+1] - C.yi[i]);
			if (x1 < C.xi[i] + vt * (C.xi[i+1] - C.xi[i])) /* P.x < intersect */
				n++; 	/* a valid crossing of y1 = C.y[i] right of x1 */
		}
	}
	return (n & 1); /* 0 if even # of crossings (out of ROI) */

	#undef C
#else

	/* The winding number algorithm */

	n = 0;
	for (i=0; i<current_roi.numvertices-1; i++) 
	{

		/* Edge from vertex i to vertex i+1 */

		if (current_roi.yi[i] <= y1)	/* Current pt above vertex i */
		{
			if (current_roi.yi[i+1] > y1)		/* upward crossing */
			{
				if (roi_is_left (i,i+1,x1,y1)>0) 	/* P left of edge */
					n++;						/* have UP intersct */
			}
		}
		else 		/* Current pt below vtx i */
		{
			if (current_roi.yi[i+1] <= y1)		/* downward crossing */						
			{
				if (roi_is_left (i,i+1, x1,y1) < 0) 	/* P right of edge */
					n--;						/* have DOWN intersct */
			}
		}
	}
	return n; /* 0 if no windings around point, out of ROI */

#endif
}



char in_roi (int x, int y)
{

	switch (current_roi.roi_mode)
	{
		case ROI_RECT:				/* Rectangular ROI */
		{
			if ( (current_roi.xi[0]<=x) && (x<=current_roi.xi[1])
				&& (current_roi.yi[0]<=y) && (y<=current_roi.yi[1]))
				return 1;
			else
				return 0;
			break;
		}
		case ROI_POLY:
		case ROI_CIRCLE:
		{
			return (winding_number (x,y)!=0);
		}
		default:
			return 1;
	}
		
}


/* Returns the bounding box */

void roi_bbox (image_type* img, int* bx1, int* by1, int* bx2, int* by2)
{

	if (!current_roi.roi_mode)	/* No ROI defaults to full image */
	{
		*bx1=0; *by1=0;
		*bx2=img->xmax-1;
		*by2=img->ymax-1;
	}
	else
	{
		*bx1 = MAX(current_roi.bbox_x1,0); *bx2=MIN(current_roi.bbox_x2,img->xmax-1);
		*by1 = MAX(current_roi.bbox_y1,0); *by2=MIN(current_roi.bbox_y2,img->ymax-1);
	}
}

/* Dump the ROI vertices, for debugging only */

void dump_roi ()
{
int bx1,by1,bx2,by2,i;

	roi_bbox (&mainimg, &bx1,&by1, &bx2, &by2);
	switch (current_roi.roi_mode)
	{
		case ROI_RECT:				/* Rectangular ROI */
		{
			dp (0, "Rectangular ROI from %dx%d to %dx%d\n",
				current_roi.xi[0], current_roi.yi[0], current_roi.xi[1], current_roi.yi[1]);
			break;
		}
		case ROI_POLY:
		case ROI_CIRCLE:
		{
			dp (0, "Vertex table for polygonal ROI\n");
			for (i=0; i<current_roi.numvertices; i++) 
			{
				dp (0, "%d:   %d   %d\n", i,current_roi.xi[i], current_roi.yi[i]);
			}
			break;

		}
	}
	

	
}




